import inspect
import logging
import os
from asyncio import CancelledError
from typing import Callable, Set, ClassVar, Type, List, Any, Hashable

import aiohttp
from scrapy import Request as ScrapyRequest, signals
from scrapy.crawler import Crawler
from scrapy.settings import Settings
from autoextract.aio import request_raw, create_session
from autoextract.aio.errors import RequestError, \
    ACCOUNT_DISABLED_ERROR_TYPE
from autoextract.aio.retry import RetryFactory
from autoextract.request import Request as AutoExtractRequest
from autoextract.stats import AggStats
from autoextract_poet.page_inputs import (
    AutoExtractProductData, AutoExtractData, AutoExtractHtml,
)
from scrapy_poet.page_input_providers import PageObjectInputProvider
from .errors import QueryError, summarize_exception
from .slot_semaphore import SlotsSemaphore
from .task_manager import TaskManager
from .utils import get_domain, get_scrapy_data_path
from .cache import AutoExtractCache, DummyCache

logger = logging.getLogger(__name__)


_TASK_MANAGER = "_autoextract_task_manager"


def get_autoextract_task_manager(crawler: Crawler) -> TaskManager:
    """
    Return the configured :class:`scrapy_autoextract.TaskManager` that controls
    AutoExtract ongoing requests.
    Handy to cancel all of them to stop fast when cancelling a spider
    """
    if not hasattr(crawler, _TASK_MANAGER):
        setattr(crawler, _TASK_MANAGER, TaskManager())
    return getattr(crawler, _TASK_MANAGER)


def _stop_if_account_disabled(exception: Exception, crawler: Crawler):
    if not isinstance(exception, RequestError):
        return

    if exception.error_data().get("type") == ACCOUNT_DISABLED_ERROR_TYPE:
        logger.info(
            "'Account disabled' request error received. Shutting down the spider"
        )
        crawler.engine.close_spider(crawler.spider, "account_disabled")


def get_concurrent_requests_per_domain(settings: Settings):
    """Return the configured AutoExtract concurrent request per domain from settings"""
    limit_name = "AUTOEXTRACT_CONCURRENT_REQUESTS_PER_DOMAIN"
    concurrency = settings.getint(limit_name, -1)
    # If no AutoExtract-specific limit is provided - use the default one
    if concurrency == -1:
        limit_name = "CONCURRENT_REQUESTS_PER_DOMAIN"
        concurrency = settings.getint(limit_name)
    if concurrency < 1:
        raise ValueError(f"Invalid '{limit_name}' "
                         f"value: {concurrency}")
    return concurrency


class AutoExtractProvider(PageObjectInputProvider):
    """Provider for AutoExtract data"""
    page_type_class: ClassVar[Type]

    # pageType requested when only html is required
    page_type_class_for_html: ClassVar[AutoExtractData] = AutoExtractProductData
    html_query_attribute: ClassVar[str] = "fullHtml"

    @classmethod
    def provided_classes(cls, type_: Callable) -> bool:
        return (inspect.isclass(type_)
                and issubclass(type_, (AutoExtractData, AutoExtractHtml)))

    def __init__(self, crawler: Crawler):
        """Initialize provider storing its dependencies as attributes."""
        self.crawler = crawler
        self.settings = crawler.spider.settings
        self.stats = crawler.stats
        self.task_manager = get_autoextract_task_manager(crawler)
        self.aiohttp_session = None
        self.crawler.signals.connect(self.on_spider_closed,
                                     signal=signals.spider_closed)
        self.retries_count = self.settings.getint(
                "AUTOEXTRACT_MAX_QUERY_ERROR_RETRIES", 0)

        per_domain_concurrency = get_concurrent_requests_per_domain(self.settings)
        self.per_domain_semaphore = SlotsSemaphore(per_domain_concurrency)

        cache_filename = self.settings.get('AUTOEXTRACT_CACHE_FILENAME')
        if cache_filename:
            cache_filename = os.path.join(get_scrapy_data_path(createdir=True),
                                          cache_filename)
            self.cache = AutoExtractCache(cache_filename)
        else:
            self.cache = DummyCache()

        logger.info(
            f"AutoExtractProvider started. Retries: {self.retries_count}, "
            f"per domain concurrency: {per_domain_concurrency}, "
            f"cache: {self.cache}"
        )

    # @property to get actual aiohttp_session, instead of predefined None in `__init__`,
    # so we can use the same session instead of creating/killing new sessions for each request
    @property
    def common_request_kwargs(self):
        return dict(
            api_key=self.settings.get("AUTOEXTRACT_USER"),
            endpoint=self.settings.get("AUTOEXTRACT_URL"),
            max_query_error_retries=self.retries_count,
            session=self.aiohttp_session
        )

    async def create_aiohttp_session(self) -> aiohttp.ClientSession:
        concurrent_connections = self.settings.getint("CONCURRENT_REQUESTS", 16)
        logger.info(
            f"AutoExtractProvider concurrent requests: {concurrent_connections}"
        )
        return create_session(connection_pool_size=concurrent_connections)

    async def on_spider_closed(self):
        if self.aiohttp_session:
            await self.aiohttp_session.close()
        self.cache.close()

    def create_retry_wrapper(self):
        return RetryFactory().build()

    async def do_request_cached(self, query, *args, **kwargs):
        assert len(query) == 1  # batches are not supported
        fp = self.cache.fingerprint(query[0])
        try:
            response = self.cache[fp]
            self.stats.inc_value("autoextract/cache/hit")
        except KeyError:
            self.stats.inc_value("autoextract/cache/miss")
            response = await self.do_request(query, *args, **kwargs)
            if "error" not in response[0]:  # don't cache errors
                self.cache[fp] = response
                self.stats.inc_value("autoextract/cache/firsthand")
            else:
                self.stats.inc_value("autoextract/cache/uncacheable")
        return response

    async def do_request(self, *args, **kwargs):
        return await request_raw(*args, **kwargs)

    def pre_process_item_data(self, data) -> Any:
        """
        Hook for transforming data before any further processing, just after
        receive the content
        """
        return data

    def get_per_domain_concurrency_slot(self, request: ScrapyRequest) -> Hashable:
        """
        Return the key that will be used to identify the domain of this request.
        This key is used to modulate the per request concurrency that can be
        set using the setting `AUTOEXTRACT_CONCURRENT_REQUESTS_PER_DOMAIN`.

        By default the key is the request domain. Override it to change
        the behavior.
        """
        return get_domain(request.url)

    def get_filled_request(self,
                           request: ScrapyRequest,
                           provided_cls: AutoExtractData,
                           should_request_html: bool) -> AutoExtractRequest:
        """Return a filled request for AutoExtract"""
        ae_request = AutoExtractRequest(
            url=request.url,
            pageType=provided_cls.page_type,
        )
        if should_request_html:
            ae_request.extra = {self.html_query_attribute: True}
        return ae_request

    async def __call__(self,
                       to_provide: Set[Callable],
                       request: ScrapyRequest
                       ) -> List:
        """Make an AutoExtract request and build a Page Input of provided class
        based on API response data.
        """
        if not self.aiohttp_session:
            self.aiohttp_session = await self.create_aiohttp_session()
        is_html_required = AutoExtractHtml in to_provide
        to_provide -= {AutoExtractHtml}
        is_extraction_required = bool(to_provide)
        if is_html_required and not is_extraction_required:
            # At least one request is required to get html
            to_provide = {self.page_type_class_for_html}

        instances = []
        for idx, provided_cls in enumerate(to_provide):
            if not issubclass(provided_cls, AutoExtractData):
                raise RuntimeError(
                    f"Unexpected {provided_cls} requested. Probably a bug in the provider "
                    "or in scrapy-poet itself")
            request_stats = AggStats()
            is_first_request = idx == 0
            page_type = provided_cls.page_type

            def inc_stats(suffix, value=1, both=False):
                self.stats.inc_value(f"autoextract/total{suffix}", value)
                if both:
                    self.stats.inc_value(f"autoextract/{page_type}{suffix}", value)

            try:
                # html is requested only a single time to save resources
                should_request_html = is_html_required and is_first_request
                slot = self.get_per_domain_concurrency_slot(request)
                ae_request = self.get_filled_request(
                    request,
                    provided_cls,
                    should_request_html
                )
                # When providing same-name arguments in both call and `__init__`
                # this implementation will not cause any errors (while name=value implementation would),
                # so predefined `__init__` arguments would override the same arguments in the call
                awaitable = self.do_request_cached(**{
                    'query': [ae_request],
                    'agg_stats': request_stats,
                    **self.common_request_kwargs,
                })
                awaitable = self.per_domain_semaphore.run(awaitable, slot)
                response = await self.task_manager.run(awaitable)
                data = response[0]
                data = self.pre_process_item_data(data)
                if "error" in data:
                    raise QueryError(data["query"], data["error"])

            except CancelledError:
                inc_stats("/pages/cancelled", both=True)
                raise
            except Exception as e:
                inc_stats("/pages/errors", both=True)
                inc_stats(f"/pages/errors{summarize_exception(e)}")
                _stop_if_account_disabled(e, self.crawler)
                raise
            finally:
                inc_stats("/pages/count", both=True)
                inc_stats("/attempts/count", request_stats.n_attempts)
                inc_stats("/attempts/billable", request_stats.n_billable_query_responses)

            if should_request_html:
                instances.append(AutoExtractHtml(url=data[page_type]['url'],
                                                 html=data['html']))
                inc_stats("/pages/html", both=True)

            if is_extraction_required:
                without_html = {k: v for k, v in data.items() if k != "html"}
                instances.append(provided_cls(data=without_html))

            inc_stats("/pages/success", both=True)

        return instances



import json
from functools import wraps

from pytest_twisted import inlineCallbacks

from autoextract.aio import RequestError
from scrapy.statscollectors import StatsCollector
from scrapy.utils.defer import maybeDeferred_coro


def async_test(f):
    """Allow running asyncio tests on the presence of pytest_twisted"""
    @inlineCallbacks
    @wraps(f)
    def fn(*args, **kwargs):
        yield maybeDeferred_coro(f, *args, **kwargs)
    return fn


def request_error(payload) -> RequestError:
    if payload is not None and not isinstance(payload, bytes):
        payload = json.dumps(payload).encode("utf-8")
    return RequestError(
        request_info=None,
        history=None,
        status=404,
        message="",
        headers=None,
        response_content=payload,
    )


def assert_stats(stats: StatsCollector, expected: dict):
    for k, expected_val in expected.items():
        actual_val = stats.get_value(k, 0)
        assert actual_val == expected_val, \
            f"key: '{k}', value: {actual_val}, expected: {expected_val}"



import sys

collect_ignore = []
if sys.version_info[:2] < (3, 7):
    # AutoExtract providers requires Python 3.7+
    collect_ignore.append("test_providers.py")
    collect_ignore.append("test_task_manager.py")
    collect_ignore.append("test_slot_semaphore.py")


from autoextract.aio.errors import DomainOccupied


class QueryError(Exception):

    def __init__(self, query: dict, message: str):
        self.query = query
        self.message = message

    def __str__(self):
        return f"QueryError: message='{self.message}', query={self.query}"


def summarize_exception(exc):
    """
    Provides a text that represents the exception. To be used in stats, so
    produced text shouldn't be too diverse.
    """
    if isinstance(exc, QueryError):
        msg = exc.message
        if DomainOccupied.from_message(msg):
            # Removes variability from domain occupy messages
            msg = "domain occupied"
        msg = f"/query/{msg}"
    else:
        msg = f"/rest/{exc.__class__.__name__}"
    return msg



import pytest

from scrapy_autoextract.errors import QueryError, summarize_exception


def test_query_error():
    exc = QueryError({"foo": "bar"}, "sample error")
    assert str(exc) == "QueryError: message='sample error', query={'foo': 'bar'}"


@pytest.mark.parametrize("exception, message", [
    (QueryError({}, "domain xyz is occupied, please retry in 2.2 seconds"),
     "/query/domain occupied"),
    (QueryError({}, "Another thing"),
     "/query/Another thing"),
    (ValueError("Value Error"), "/rest/ValueError"),
    (TypeError("Type Error"), "/rest/TypeError"),
])
def test_summarize_exception(exception, message):
    assert summarize_exception(exception) == message


import time
import json
import logging

import scrapy
from scrapy import signals
from scrapy.http import Headers, HtmlResponse
from scrapy.utils.python import global_object_name
from scrapy.exceptions import IgnoreRequest, DropItem
from w3lib.http import basic_auth_header

from .__version__ import __version__

logger = logging.getLogger(__name__)

AUTOEXTRACT_META_KEY = '_autoextract_processed'
USER_AGENT = 'scrapy-autoextract/{} scrapy/{}'.format(__version__, scrapy.__version__)
MAX_ERROR_BODY = 2000


class AutoExtractError(Exception):
    pass


class AutoExtractConfigError(Exception):
    pass


class SlotPolicy(object):
    PER_DOMAIN = 'per_domain'
    SINGLE_SLOT = 'single_slot'
    SCRAPY_DEFAULT = 'scrapy_default'


class AutoExtractMiddleware(object):
    """
    Middleware that allows a Scrapy Spider to receive Automatic Data Extraction
    results within the `response.meta`.

    The service URL can be specified with `AUTOEXTRACT_URL` in scrapy settings.
    """

    DEFAULT_URL = 'https://autoextract.scrapinghub.com/v1/extract'
    DEFAULT_TIMEOUT = 660
    DEFAULT_SLOT_POLICY = SlotPolicy.PER_DOMAIN

    def __init__(self, crawler):
        self.crawler = crawler
        self.settings = crawler.settings
        self._api_user = self.settings['AUTOEXTRACT_USER']
        self._api_pass = ''
        self.page_type = self.settings['AUTOEXTRACT_PAGE_TYPE']
        if not self.page_type:
            self.page_type = getattr(crawler.spider, 'page_type', None)
        self.timeout = max(
            self.settings.getint('AUTOEXTRACT_TIMEOUT', self.DEFAULT_TIMEOUT),
            self.settings.getint('DOWNLOAD_TIMEOUT', 180),
        )
        self.slot_policy = self.settings.get('AUTOEXTRACT_SLOT_POLICY', self.DEFAULT_SLOT_POLICY)

        self._api_url = self.settings.get('AUTOEXTRACT_URL', self.DEFAULT_URL)
        logger.info('Using AutoExtract API URL: %s', self._api_url, extra={'spider': crawler.spider})

        self.nr_resp = 0
        self.avg_latency = 0
        self.max_latency = 0
        self.total_latency = 0

    @classmethod
    def from_crawler(cls, crawler):
        if not crawler.settings.get('AUTOEXTRACT_USER'):
            raise AutoExtractConfigError('AUTOEXTRACT_USER is required')
        o = cls(crawler)
        crawler.signals.connect(o.spider_closed, signal=signals.spider_closed)
        return o

    def spider_closed(self, spider):
        if self.nr_resp > 0:
            self.autoextract_latency_stats()
            logger.info(
                'Total "%i" responses from AutoExtract, average latency=%.3f, max latency=%.3f',
                self.nr_resp,
                self.avg_latency,
                self.max_latency,
                extra={'spider': spider},
            )

    def process_request(self, request, spider):
        """
        The request will be passed to the AutoExtract server only if the request
        is explicitly enabled with `{'autoextract': {'enabled': True}}` meta.
        The page type value must be also present, either in the
        AUTOEXTRACT_PAGE_TYPE option, or in `{'autoextract': {'pageType': '...'}}` meta.
        """
        if not self._is_enabled_for_request(request):
            return

        # If the request was already processed by AutoExtract
        if request.meta.get(AUTOEXTRACT_META_KEY):
            return

        if request.method != 'GET':
            raise AutoExtractError('Only GET requests are supported by AutoExtract')

        request.meta[AUTOEXTRACT_META_KEY] = {
            'original_url': request.url,
            'timing': {
                'start_ts': time.time()
            },
        }

        # Maybe over-write the page type value from the request
        page_type = self._check_page_type(request)
        logger.debug('Process AutoExtract request for %s URL %s',
                     page_type,
                     request,
                     extra={'spider': spider})

        # Define request timeout
        request.meta['download_timeout'] = self.timeout

        # Define concurrency settings
        self._set_download_slot(request, request.meta)

        payload = {'url': request.url, 'pageType': page_type}

        # Add the extra payload, if available
        extra_payload = self._get_meta_name(request, 'extra')
        if extra_payload:
            payload.update(extra_payload)

        headers = Headers({
            'Content-Type': 'application/json',
            'User-Agent': USER_AGENT,
            'Authorization': basic_auth_header(self._api_user, self._api_pass)
        })
        # Update the headers, if provided
        extra_headers = self._get_meta_name(request, 'headers')
        if extra_headers:
            headers.update(extra_headers)

        new_request = request.replace(
            url=self._api_url,
            method='POST',
            headers=headers,
            body=json.dumps([payload], sort_keys=True),
        )

        self.inc_metric('autoextract/request_count')
        return new_request

    def process_response(self, request, response, spider):
        if not self._is_enabled_for_request(request):
            return response

        # If the request was never processed by AutoExtract
        if not request.meta.get(AUTOEXTRACT_META_KEY):
            return response

        url = request.meta[AUTOEXTRACT_META_KEY]['original_url']
        body = response.body.decode('utf8')

        try:
            response_object = json.loads(body)
        except Exception:
            self.inc_metric('autoextract/errors/json_decode')
            self._log_debug_error(response, body)
            raise AutoExtractError('Cannot parse JSON response from AutoExtract'
                                   ' for {}: {}'.format(url, response.body[:MAX_ERROR_BODY]))

        if response.status != 200:
            self.inc_metric('autoextract/errors/response_error/{}'.format(response.status))
            self._log_debug_error(response, body)
            raise AutoExtractError('Received error from AutoExtract for '
                                   '{}: {}'.format(url, response_object))

        result = None
        if isinstance(response_object, list):
            result = response_object[0]
        else:
            self.inc_metric('autoextract/errors/type_error')
            self._log_debug_error(response, body)
            raise AutoExtractError('Received invalid response from AutoExtract for '
                                   '{}: {}'.format(url, response_object))

        if result.get('error'):
            self.inc_metric('autoextract/errors/result_error')
            self._log_debug_error(response, body)
            raise AutoExtractError('Received error from AutoExtract for '
                                   '{}: {}'.format(url, result["error"]))

        stop_time = time.time()
        latency = stop_time - request.meta[AUTOEXTRACT_META_KEY]['timing']['start_ts']

        self.nr_resp += 1
        self.total_latency += latency
        self.max_latency = max(self.max_latency, latency)
        self.avg_latency = float(self.total_latency) / self.nr_resp
        self.autoextract_latency_stats()

        autoextract = request.meta.pop(AUTOEXTRACT_META_KEY)
        # Middleware-level timings
        autoextract['timing'].update({'end_ts': stop_time, 'latency': latency})

        page_type = self._check_page_type(request)
        logger.debug('AutoExtract latency for %s URL %s was %.3fs',
                     page_type,
                     url,
                     latency,
                     extra={'spider': spider})

        # The AutoExtract processed item is added here
        autoextract[page_type] = result.get(page_type) or {}
        request.meta['autoextract'] = autoextract
        page_html = result.pop('html', '<body></body>')

        return HtmlResponse(
            url,
            request=request,
            body=page_html.encode('utf-8'),
            encoding='utf-8',
        )

    def process_exception(self, request, exception, spider):
        if isinstance(exception, (IgnoreRequest, DropItem)):
            return
        if not self._is_enabled_for_request(request):
            return

        autoextract = request.meta.pop(AUTOEXTRACT_META_KEY)
        stop_time = time.time()
        latency = time.time() - autoextract['timing']['start_ts']
        autoextract['timing'].update({'end_ts': stop_time, 'latency': latency})

        # Make sure to log all unknown failures
        logger.warning('AutoExtract failure after %.3fs for %s: %s',
                       latency,
                       autoextract['original_url'],
                       repr(exception),
                       extra={'spider': spider})

        request.meta['autoextract'] = autoextract
        ex_class = global_object_name(exception.__class__)
        self.inc_metric('autoextract/errors/total_count', spider=spider)
        self.inc_metric('autoextract/errors/type_count/%s' % ex_class, spider=spider)

    def _is_enabled_for_request(self, request):
        if 'autoextract' not in request.meta:
            return False
        if not request.meta.get('autoextract', {}).get('enabled'):
            return False
        return True

    def _check_page_type(self, request):
        # Use pageType value from the request.meta['autoextract']
        # and fallback to the value from middleware
        page_type = request.meta.get('autoextract', {}).get('pageType', self.page_type)
        if not page_type or not isinstance(page_type, str):
            raise AutoExtractConfigError('Invalid pageType value: {}'.format(page_type))
        return page_type

    def _set_download_slot(self, request, meta):
        if self.slot_policy == SlotPolicy.PER_DOMAIN:
            # Group requests by domain to respect download
            # delays and concurrency options
            slot = self.crawler.engine.downloader._get_slot_key(request, None)
            meta['download_slot'] = slot
        elif self.slot_policy == SlotPolicy.SINGLE_SLOT:
            # Use a single slot for all AutoExtract requests
            meta['download_slot'] = '__AutoExtract__'
        # Else, use standard Scrapy concurrency setup

    def _get_meta_name(self, request, name):
        extra_data = None
        if request.meta['autoextract'].get(name):
            if not isinstance(request.meta['autoextract'][name], dict):
                raise AutoExtractError('Invalid type for "{}" meta'.format(name))
            extra_data = request.meta['autoextract'][name]
        return extra_data

    def inc_metric(self, key, **kwargs):
        self.crawler.stats.inc_value(key, **kwargs)

    def set_metric(self, key, value):
        self.crawler.stats.set_value(key, value)

    def _log_debug_error(self, response, body):
        if len(body) > MAX_ERROR_BODY:
            half_body = MAX_ERROR_BODY // 2
            body = body[:half_body] + ' [...] ' + body[-half_body:]
        logger.debug('AutoExtract response status=%i  headers=%s  content=%s', response.status,
                     response.headers.to_unicode_dict(), body)

    def autoextract_latency_stats(self):
        self.set_metric('autoextract/response_count', self.nr_resp)
        self.set_metric('autoextract/response_avg_latency', self.avg_latency)
        self.set_metric('autoextract/response_max_latency', self.max_latency)



import asyncio
import logging
import signal
from asyncio import CancelledError
from typing import Awaitable, Set, TypeVar

from scrapy.utils.ossignal import signal_names

logger = logging.getLogger(__name__)

T = TypeVar("T")


class TaskManager:
    """
    A class that runs async functions keeping track of those that are running.
    The tasks can be cancelled, and actually they are cancelled
    when ``signal.SIGINT`` is received.

    Note that cancelled tasks will receive exception ``CancelledError``

    Any task submitted after receive the cancelling order will raise also
    ``CancelledError``

    The created task managers will live during the entire lifecycle of the process
    because the signal handler needs the reference to cancel tasks on
    shutdown.

    Example usage::

        async def my_task(sleep):
            await asyncio.sleep(sleep)
            return "Finished"

        manager = TaskManager()
        for i in range(10):
            print(await manager.run(my_task(i)))

        await asyncio.sleep(5)
        manager.cancel_all()
    """

    def __init__(self, cancel_on_signals={signal.SIGINT, signal.SIGTERM}):
        self.running_tasks: Set[asyncio.Task] = set()
        self.cancelled = False
        for sig in cancel_on_signals:
            self._install_signal_handler(sig)

    def _install_signal_handler(self, sig):
        """Installs the signal handler for cancellation, respecting existing handler"""
        old_handler = signal.getsignal(sig)

        def new_handler(*args, **kwargs):
            # Invoke old handler if any
            exception = None
            if callable(old_handler):
                try:
                    old_handler(*args, **kwargs)
                except BaseException as e:
                    exception = e
            self._cancel_on_signal(*args, **kwargs)
            if exception:
                raise exception

        signal.signal(sig, new_handler)

    async def run(self, awaitable: Awaitable[T]) -> T:
        """Run a task"""
        task = asyncio.create_task(awaitable)
        self.running_tasks.add(task)
        try:
            # This avoids a potential race condition
            if self.cancelled:
                raise CancelledError()
            return await task
        finally:
            self.running_tasks.remove(task)

    def _cancel_on_signal(self, sig, frame):
        logger.info(
            f"Received {signal_names[sig]}. Cancelling {len(self.running_tasks)} running tasks"
        )
        self.cancel_all()

    def cancel_all(self):
        """
        Cancel all running tasks and any future submitted task
        """
        self.cancelled = True
        for task in self.running_tasks:
            task.cancel()

    def __len__(self):
        return len(self.running_tasks)



import asyncio
import copy
import os
import signal
from asyncio import CancelledError
from signal import SIGINT

import pytest
from pytest_mock import MockerFixture
from pytest_twisted import inlineCallbacks
from twisted.internet.defer import Deferred

from autoextract.aio.errors import ACCOUNT_DISABLED_ERROR_TYPE
from autoextract.stats import AggStats
from autoextract_poet import (
    AutoExtractArticleData, AutoExtractProductData, AutoExtractHtml)
from tests.utils import assert_stats, request_error, async_test
from autoextract_poet.page_inputs import AutoExtractData
from scrapy import Spider
from scrapy.crawler import Crawler
from scrapy_autoextract.providers import (
    AutoExtractProvider, _stop_if_account_disabled,
)
from scrapy_autoextract.errors import QueryError
from scrapy_poet.injection import get_injector_for_testing, get_response_for_testing

DATA_INPUTS = (
    AutoExtractArticleData,
    AutoExtractProductData,
)


def test_stop_spider_on_account_disabled(mocker: MockerFixture):
    class Engine:
        close_spider = mocker.Mock()

    class _EmptySpider(Spider):
        name = "empty_spider"

    crawler = Crawler(_EmptySpider)
    crawler.engine = Engine()
    spider = _EmptySpider.from_crawler(crawler)
    crawler.spider = spider

    _stop_if_account_disabled(Exception(), crawler)
    spider.crawler.engine.close_spider.assert_not_called()

    re = request_error({"type": "whatever"})
    _stop_if_account_disabled(re, crawler)
    spider.crawler.engine.close_spider.assert_not_called()

    re = request_error({"type": ACCOUNT_DISABLED_ERROR_TYPE})
    _stop_if_account_disabled(re, crawler)
    spider.crawler.engine.close_spider.assert_called_with(spider, "account_disabled")


class TestProviders:

    @inlineCallbacks
    @pytest.mark.parametrize("provided_cls", DATA_INPUTS)
    def test_providers(self, provided_cls: AutoExtractProductData):
        page_type = provided_cls.page_type
        url, html = "http://example.com", "html_content"
        data_wo_html = {page_type: {"url": url}}
        data = {page_type: {"url": url}, "html": html}
        provider_wrapper = []

        class Provider(AutoExtractProvider):
            async def do_request(self, *args, agg_stats: AggStats, **kwargs):
                assert provider.aiohttp_session.connector.limit == 2020
                agg_stats.n_attempts += 3
                agg_stats.n_billable_query_responses += 2
                assert kwargs['api_key'] == "key"
                assert kwargs['endpoint'] == "url"
                assert kwargs['max_query_error_retries'] == 31415
                return [copy.deepcopy(data)]

        def callback(item: provided_cls):
            pass

        def callback_with_html(item: provided_cls, html: AutoExtractHtml):
            pass

        def callback_only_html(html: AutoExtractHtml):
            pass

        settings = {
            "AUTOEXTRACT_USER": "key",
            "AUTOEXTRACT_URL": "url",
            "AUTOEXTRACT_MAX_QUERY_ERROR_RETRIES": 31415,
            "CONCURRENT_REQUESTS": 2020,
            "CONCURRENT_REQUESTS_PER_DOMAIN": 1980,
        }
        injector = get_injector_for_testing({Provider: 500}, settings)
        stats = injector.crawler.stats
        provider = injector.providers[-1]
        provider_wrapper.append(provider)
        assert provider.per_domain_semaphore.concurrency_per_slot == 1980

        #  - No HTML requested case -

        response = get_response_for_testing(callback)
        kwargs = yield injector.build_callback_dependencies(response.request,
                                                          response)
        assert kwargs["item"].data == data_wo_html
        assert type(kwargs["item"]) is provided_cls
        expected_stats = {
            'autoextract/total/pages/count': 1,
            'autoextract/total/pages/success': 1,
            'autoextract/total/attempts/count': 3,
            'autoextract/total/attempts/billable': 2,
            f'autoextract/{page_type}/pages/count': 1,
            f'autoextract/{page_type}/pages/success': 1
        }
        assert_stats(stats, expected_stats)

        #  - Both HTML and item requested case -

        response = get_response_for_testing(callback_with_html)
        kwargs = yield injector.build_callback_dependencies(response.request,
                                                          response)
        item, html_response = kwargs["item"], kwargs["html"]
        assert item.data == data_wo_html
        assert type(item) is provided_cls
        assert (html_response.url, html_response.html) == (url, html)
        assert type(html_response) is AutoExtractHtml
        expected_stats = {
            'autoextract/total/pages/count': 2,
            'autoextract/total/pages/success': 2,
            'autoextract/total/pages/html': 1,
            'autoextract/total/attempts/count': 6,
            'autoextract/total/attempts/billable': 4,
            f'autoextract/{page_type}/pages/count': 2,
            f'autoextract/{page_type}/pages/success': 2,
            f'autoextract/{page_type}/pages/html': 1,
        }
        assert_stats(stats, expected_stats)

        #  - Only HTML is requested case -

        injector.providers[0].page_type_class_for_html = provided_cls
        response = get_response_for_testing(callback_only_html)
        kwargs = yield injector.build_callback_dependencies(response.request,
                                                          response)
        assert "item" not in kwargs
        html_response = kwargs["html"]
        assert (html_response.url, html_response.html) == (url, html)
        assert type(html_response) is AutoExtractHtml
        expected_stats = {
            'autoextract/total/pages/count': 3,
            'autoextract/total/pages/success': 3,
            'autoextract/total/pages/html': 2,
            'autoextract/total/attempts/count': 9,
            'autoextract/total/attempts/billable': 6,
            f'autoextract/{page_type}/pages/count': 3,
            f'autoextract/{page_type}/pages/success': 3,
            f'autoextract/{page_type}/pages/html': 2,
        }
        assert_stats(stats, expected_stats)


    @inlineCallbacks
    @pytest.mark.parametrize("provided_cls", DATA_INPUTS)
    def test_on_query_error(self, provided_cls: AutoExtractData):
        page_type = provided_cls.page_type
        data = {"query": "The query", "error": "Download error"}

        class Provider(AutoExtractProvider):
            async def do_request(self, *args, agg_stats: AggStats, **kwargs):
                agg_stats.n_attempts += 3
                agg_stats.n_billable_query_responses += 2
                return [data]

        def callback(item: provided_cls):
            pass

        injector = get_injector_for_testing({Provider: 500})
        response = get_response_for_testing(callback)
        with pytest.raises(QueryError) as exinf:
            yield injector.build_callback_dependencies(response.request, response)
        stats = injector.crawler.stats
        expected = {
            f'autoextract/{page_type}/pages/count': 1,
            f'autoextract/{page_type}/pages/errors': 1,
            'autoextract/total/attempts/count': 3,
            'autoextract/total/attempts/billable': 2,
            'autoextract/total/pages/count': 1,
            'autoextract/total/pages/errors': 1,
            'autoextract/total/pages/errors/query/Download error': 1
        }
        assert_stats(stats, expected)
        assert "Download error" in str(exinf.value)
        assert "The query" in str(exinf.value)

    @inlineCallbacks
    @pytest.mark.parametrize("provided_cls", DATA_INPUTS)
    def test_on_exception(self, provided_cls: AutoExtractData):

        class Provider(AutoExtractProvider):
            async def do_request(self, *args, agg_stats: AggStats, **kwargs):
                agg_stats.n_attempts += 3
                agg_stats.n_billable_query_responses += 2
                raise Exception()

        def callback(item: provided_cls):
            pass

        page_type = provided_cls.page_type
        injector = get_injector_for_testing({Provider: 500})
        response = get_response_for_testing(callback)
        with pytest.raises(Exception) as exinf:
            yield injector.build_callback_dependencies(response.request, response)
        stats = injector.crawler.stats
        expected = {
            f'autoextract/{page_type}/pages/count': 1,
            f'autoextract/{page_type}/pages/errors': 1,
            'autoextract/total/attempts/count': 3,
            'autoextract/total/attempts/billable': 2,
            'autoextract/total/pages/count': 1,
            'autoextract/total/pages/errors': 1,
            'autoextract/total/pages/errors/rest/Exception': 1
        }
        assert_stats(stats, expected)

    @async_test
    @pytest.mark.parametrize("provided_cls", DATA_INPUTS)
    async def test_on_cancellation(self, provided_cls: AutoExtractProductData):
        old_handler = signal.getsignal(SIGINT)
        signal.signal(SIGINT, lambda x, y: None)
        try:
            lock = asyncio.Lock()
            await lock.acquire()

            class Provider(AutoExtractProvider):
                async def do_request(self, *args, agg_stats: AggStats, **kwargs):
                    await lock.acquire()

            def callback(item: provided_cls):
                pass

            injector = get_injector_for_testing({Provider: 500})
            stats = injector.crawler.stats
            response = get_response_for_testing(callback)
            deferred = injector.build_callback_dependencies(response.request,
                                                              response)
            build_callbacks_future = Deferred.asFuture(deferred, asyncio.get_event_loop())

            async def cancel_after(sleep):
                await asyncio.sleep(sleep)
                pid = os.getpid()
                try:
                    os.kill(pid, SIGINT)
                except KeyboardInterrupt:
                    # As an effect of the SIGINT killing the process might receive
                    # here a KeyboardInterrupt exception. This is Ok.
                    pass
                return CancelledError()

            result = await asyncio.gather(
                build_callbacks_future, cancel_after(0.05), return_exceptions=True
            )
            assert all([isinstance(r, CancelledError) for r in result])

            page_type = provided_cls.page_type
            expected_stats = {
                'autoextract/total/pages/count': 1,
                'autoextract/total/pages/cancelled': 1,
                'autoextract/total/pages/errors': 0,
                f'autoextract/{page_type}/pages/count': 1,
                f'autoextract/{page_type}/pages/cancelled': 1,
                f'autoextract/{page_type}/pages/errors': 0,
            }
            assert_stats(stats, expected_stats)

        finally:
            signal.signal(SIGINT, old_handler)



from asyncio import BoundedSemaphore
from typing import Dict, Hashable, Awaitable, TypeVar

import attr

T = TypeVar("T")


class SlotsSemaphore:
    """
    A synchronization primitive that keeps one semaphore per living slot.
    Useful for limiting concurrency within each slot without limiting concurrency
    between the slots.

    Internally, a dictionary is keeping a semaphore per each slot. When no
    more tasks pending or running for this slot, the semaphore for the slot
    is freed, so that the dictionary does not ever grow with slots.

    Example usage::

        sem = SlotsSemaphore(2)

        async def task(url):
            async with sem.use_slot(get_domain(url)):
                await fetch(url)

        results = await asyncio.gather(*[task(url) for url in urls])

    Or alternatively::

        sem = SlotsSemaphore(2)
        results = await asyncio.gather(*[sem.run(fetch(url), get_domain(url))
                                         for url in urls])


    Ideas for the future:
    - To avoid excessive creation and destruction of semaphores,
      the policy regarding when to evict an empty semaphore could be changed.
      Using a LRU cache instead of a dict might do the job.
    """

    def __init__(self, concurrency_per_slot: int):
        """
        :param concurrency_per_slot: maximum number of tasks that can run
                                     concurrently per each slot
        """
        self.concurrency_per_slot = concurrency_per_slot
        self.slots: Dict[Hashable, _SlotMeta] = {}

    def use_slot(self, slot: Hashable):
        """
        Return an asynchronous context manager to wrap your async code so
        that it respect the concurrency limits
        """
        return _SlotSemaphore(self, slot)

    async def run(self, awaitable: Awaitable[T], slot: Hashable) -> T:
        """Run the given awaitable respecting the concurrency limits"""
        async with self.use_slot(slot):
            return await awaitable

    def __str__(self):
        slots_str = "empty"
        if self.slots:
            counts_str = ",".join(
                str(info.registered_tasks) for info in self.slots.values()
            )
            slots_str = f"{len(self.slots)} slots. Tasks counts {counts_str}"
        return f"SlotsSemaphore({self.concurrency_per_slot}) {slots_str}"


@attr.s(auto_attribs=True)
class _SlotMeta:
    semaphore: BoundedSemaphore
    registered_tasks: int


class _SlotSemaphore:
    """
    Keys:
    - When no slot is present in the slots dictionary, a new one is created.
    - The number of registered tasks is tracked
    - When the number of registered tasks reaches 0 after releasing, the slot
      is removed
    """

    def __init__(self, parent: SlotsSemaphore, slot: Hashable):
        self.parent = parent
        self.slot = slot

    async def __aenter__(self):
        if self.slot not in self.parent.slots:
            self.parent.slots[self.slot] = _SlotMeta(
                BoundedSemaphore(self.parent.concurrency_per_slot), registered_tasks=0
            )
        slot_info = self.parent.slots[self.slot]
        slot_info.registered_tasks += 1
        await slot_info.semaphore.acquire()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        slot_info = self.parent.slots[self.slot]
        slot_info.semaphore.release()
        slot_info.registered_tasks -= 1
        if slot_info.registered_tasks == 0:
            del self.parent.slots[self.slot]



import asyncio
import os
import signal
from asyncio import CancelledError
from collections import Counter

import pytest
from scrapy_autoextract.task_manager import TaskManager
from tests.utils import async_test


class _TasksBench:
    def __init__(self):
        self.lock = asyncio.Lock()
        self.manager = TaskManager(cancel_on_signals={signal.SIGUSR1})
        self.n = 10

    async def identity(self, value):
        """Waits on the lock and then just return value"""
        assert len(self.manager.running_tasks) == self.n
        async with self.lock:
            return value

    async def release_after(self, sleep):
        """Releases the lock after some time"""
        await asyncio.sleep(sleep)
        assert len(self.manager.running_tasks) == self.n
        self.lock.release()
        return 0

    async def cancel_after(self, sleep):
        """Cancel all tasks after some time"""
        await asyncio.sleep(sleep)
        assert len(self.manager.running_tasks) == self.n
        self.manager.cancel_all()
        return CancelledError()

    async def kill_after(self, sleep):
        """Send SIGINT signal after some time"""
        await asyncio.sleep(sleep)
        assert len(self.manager.running_tasks) == self.n
        pid = os.getpid()
        os.kill(pid, signal.SIGUSR1)
        return CancelledError()


@pytest.fixture()
def tasks_bench():
    return _TasksBench()


class TestTaskManager:

    @async_test
    async def test_run(self, tasks_bench):
        """Run 10 tasks and waits for it"""
        await tasks_bench.lock.acquire()
        tasks = [
            tasks_bench.manager.run(tasks_bench.identity(i))
            for i in range(tasks_bench.n)
        ] + [tasks_bench.release_after(0.05)]
        result = await asyncio.gather(*tasks)
        assert sum(result) == sum(range(tasks_bench.n))
        assert len(tasks_bench.manager.running_tasks) == 0

    @async_test
    async def test_cancel_all(self, tasks_bench):
        """Run 10 tasks that waits on lock, and at some point all are cancelled"""
        await tasks_bench.lock.acquire()
        tasks = [
            tasks_bench.manager.run(tasks_bench.identity(i))
            for i in range(tasks_bench.n)
        ] + [tasks_bench.cancel_after(0.05)]
        result = await asyncio.gather(*tasks, return_exceptions=True)
        assert all(isinstance(r, CancelledError) for r in result)
        assert len(tasks_bench.manager.running_tasks) == 0

        # Submitted tasks after cancellation must be cancelled as well
        with pytest.raises(CancelledError):
            await tasks_bench.manager.run(tasks_bench.identity(1))

    @async_test
    async def test_signal_cancelation(self):
        """ Run 10 tasks that waits on lock, and at some point all are cancelled by a signal"""
        old_signal_handler_called = []

        def old_signal_handler(*args):
            old_signal_handler_called.append(True)
            raise ValueError(
                """
                The new handler should executed even if an error in raised by the old one.
                Also, the exception from the old handler should be raised.
                """
            )

        signal.signal(signal.SIGUSR1, old_signal_handler)

        tasks_bench = _TasksBench()
        await tasks_bench.lock.acquire()
        tasks = [
            tasks_bench.manager.run(tasks_bench.identity(i))
            for i in range(tasks_bench.n)
        ] + [tasks_bench.kill_after(0.05)]
        result = await asyncio.gather(*tasks, return_exceptions=True)
        type_count = Counter(type(r) for r in result)
        assert type_count.keys() == {CancelledError, ValueError}
        assert type_count[CancelledError] == 10
        assert len(tasks_bench.manager.running_tasks) == 0
        assert old_signal_handler_called

        # Submitted tasks after cancellation must be cancelled as well
        with pytest.raises(CancelledError):
            await tasks_bench.manager.run(tasks_bench.identity(1))



import asyncio
from collections import Counter

from scrapy_autoextract.slot_semaphore import SlotsSemaphore

from .utils import async_test


class TaskSimulator:
    """
    Executes dummy tasks using ``SlotSemaphore``. Useful for unit testing.
    """

    def __init__(self, slot_concurrency):
        self.slot_concurrency = slot_concurrency

        self.total_runned_tasks = 0
        self.task_id_generator = 0
        self.slot_sem = SlotsSemaphore(self.slot_concurrency)
        self.registered_tasks = Counter()
        self.running_tasks = Counter()
        self.max_slots_in_parallel = 0

    async def task(self, slot, sleep, duration):
        """
        Executed a dummy task for the given slot.
        :param slot: The slot
        :param sleep: sleep time before acquiring an semaphore executing the time
        :param duration: duration of the task
        """
        self.task_id_generator = self.task_id_generator + 1
        task_id = self.task_id_generator

        def log(text):
            print(
                f"[{task_id:3}] {text}. Parallelism {self.running_tasks[slot]}. Registered tasks {self.registered_tasks[slot]}"
            )

        await asyncio.sleep(sleep)

        assert len(str(self.slot_sem)) > 0
        self.registered_tasks[slot] += 1
        log("Acquiring")
        await self.slot_sem.run(self.run_task(duration, log, slot), slot)
        self.running_tasks[slot] -= 1
        self.total_runned_tasks += 1
        self.registered_tasks[slot] -= 1
        assert len(str(self.slot_sem)) > 0
        log("Finished")

    async def run_task(self, duration, log, slot):
        self.running_tasks[slot] += 1
        log("Running")
        assert (
                self.running_tasks[slot] > 0
                and self.running_tasks[slot] <= self.slot_concurrency
        )
        assert len(str(self.slot_sem)) > 0
        self.max_slots_in_parallel = max(
            self.max_slots_in_parallel,
            sum([1 for tasks in self.running_tasks.values() if tasks]),
        )
        await asyncio.sleep(duration)


@async_test
async def test_slot_semaphore_simple_case():
    parallelism = 3
    sim = TaskSimulator(parallelism)
    duration = 0.05
    await asyncio.gather(
        sim.task(1, 0, duration),
        sim.task(1, 0, duration),
        sim.task(1, 0, duration),
        sim.task(1, 0, duration),
    )
    assert sim.total_runned_tasks == 4
    assert sim.slot_sem.slots == {}


@async_test
async def test_slot_semaphore():
    """
    Simulate the execution of concurrent tasks respecting a given
    parallelism per slot.

    The time is divided in a fixed number of ticks. For each tick,
    N tasks can be executed. At the first tick, all tasks are
    scheduled. For the second tick, only the par tasks are executed: 0, 2, 4, ..
    For the third one, only those divisible by 3. And so on. And this happens
    cyclically each N ticks.

    The same is happening for several slots. For the first slot, there is
    no delay between tasks, so all tasks for a particular tick are scheduled
    for the same time. But for slots higher than 0 some small delay is introduced
    to misalign them.

    So in conclusion, the following is tested:
    - running more tasks than the allowed parallelism, but respecting
      the parallelism
    - running tasks for different slots
    """
    parallelism = 3
    concurrency = 2 * parallelism
    slots = 3
    clock_tick = 0.025
    ticks = 30

    tasks = []
    sim = TaskSimulator(parallelism)
    for tick in range(ticks):
        for task in range(concurrency):
            for slot in range(slots):
                if task % ((tick % concurrency) + 1) == 0:
                    start_time = tick * clock_tick
                    # Introducing some delays. No delays in slot 0.
                    # Higher delay as the slot is higher
                    # Again, no delay is task is 0.
                    # Higher delays is the task is higher
                    # Delay always bounded by clock_tick
                    delay_multiplier = (slot * task) / (slots * concurrency)
                    delay = delay_multiplier * clock_tick
                    start_time += delay
                    tasks.append(sim.task(slot, start_time, clock_tick))

    await asyncio.gather(*tasks)
    assert sim.total_runned_tasks == len(tasks)
    assert sim.slot_sem.slots == {}
    assert sim.max_slots_in_parallel == parallelism



import os

from tldextract import tldextract
from scrapy.utils.project import inside_project, project_data_dir


def get_domain(url):
    """
    Return the domain without any subdomain

    >>> get_domain("http://blog.example.com")
    'example.com'
    >>> get_domain("http://www.example.com")
    'example.com'
    >>> get_domain("http://deeper.blog.example.co.uk")
    'example.co.uk'
    """
    return ".".join(tldextract.extract(url)[-2:])


def get_scrapy_data_path(createdir=True):
    """ Return a path to a folder where Scrapy is storing data.
    Usually that's a .scrapy folder inside the project.
    """
    # This code is extracted from scrapy.utils.project.data_path function,
    # which does too many things.
    path = project_data_dir() if inside_project() else ".scrapy"
    if createdir:
        os.makedirs(path, exist_ok=True)
    return path



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
from setuptools import setup, find_packages

NAME = 'scrapy-autoextract'


def get_version():
    about = {}
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, NAME.replace('-', '_'), '__version__.py')) as f:
        exec(f.read(), about)
    return about['__version__']


setup(
    name=NAME,
    version=get_version(),
    author='Scrapinghub Inc',
    author_email='info@scrapinghub.com',
    maintainer='Scrapinghub Inc',
    maintainer_email='info@scrapinghub.com',
    description='Scrapinghub AutoExtract API integration for Scrapy',
    long_description=open('README.rst').read(),
    url='https://github.com/scrapinghub/scrapy-autoextract',
    packages=find_packages(),
    install_requires=[
        'autoextract-poet>=0.2.1',
        'zyte-autoextract>=0.7.0',
        'scrapy-poet>=0.2.0',
        'aiohttp',
        'tldextract',
        'sqlitedict>=1.7.0',
    ],
    keywords='scrapy autoextract middleware',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Framework :: Scrapy',
    ],
)



import json
import pytest
from copy import deepcopy

from w3lib.http import basic_auth_header
from scrapy.http import Request, Response
from scrapy.spiders import Spider
from scrapy.utils.test import get_crawler

from scrapy_autoextract import AutoExtractMiddleware
from scrapy_autoextract.middlewares import AutoExtractError, AutoExtractConfigError

AUTOX_META = {'autoextract': {'enabled': True}}

MW_SETTINGS = {
    'AUTOEXTRACT_USER': 'apikey',
    'AUTOEXTRACT_PAGE_TYPE': 'article',
}


def setup_module(module):
    global spider
    spider = Spider('spidr')


def _mock_mw(spider, settings=None):

    class MockedDownloader:
        slots = {}

        def _get_slot_key(self, a, b):
            return str(a) + str(b)

    class MockedEngine:
        downloader = MockedDownloader()
        fake_spider_closed_result = None

        def close_spider(self, spider, reason):
            self.fake_spider_closed_result = (spider, reason)

    # with `spider` instead of `type(spider)` raises an exception
    crawler = get_crawler(type(spider), settings)
    crawler.engine = MockedEngine()
    return AutoExtractMiddleware.from_crawler(crawler)


def _assert_disabled(spider, settings=None):
    mw = _mock_mw(spider, settings)
    req = Request('http://quotes.toscrape.com', meta=AUTOX_META)
    out = mw.process_request(req, spider)
    assert out is None
    assert req.meta.get('autoextract') is None
    res = Response(req.url, request=req)
    assert mw.process_response(req, res, spider) == res


def _assert_enabled(spider,
                    settings=None,
                    url='http://quotes.toscrape.com',
                    api_url='autoextract.scrapinghub.com',
                    api_auth=basic_auth_header('apikey', '')):
    mw = _mock_mw(spider, settings)

    req = Request(url, meta=AUTOX_META)
    out = mw.process_request(req, spider)
    assert api_url in out.url
    assert out.meta['autoextract'].get('enabled')
    assert out.headers.get('Authorization') == api_auth
    assert 'User-Agent' in out.headers

    resp = Response(out.url, request=out, body=b'[{}]')
    proc = mw.process_response(out, resp, spider)
    assert proc.meta['autoextract'].get('original_url') == url
    assert isinstance(proc.meta['autoextract'].get('article'), dict)


def test_bad_config():
    with pytest.raises(AutoExtractConfigError):
        _assert_disabled(spider, {})
    with pytest.raises(AutoExtractConfigError):
        _assert_disabled(spider, {'AUTOEXTRACT_USER': 'apikey'})


def test_disabled():
    mw = _mock_mw(spider, MW_SETTINGS)
    req = Request('http://quotes.toscrape.com', meta={'autoextract': {}})
    out = mw.process_request(req, spider)
    assert out is None


def test_enabled():
    _assert_enabled(spider, MW_SETTINGS)


def test_request_error():
    mw = _mock_mw(spider, MW_SETTINGS)
    req = Request('http://quotes.toscrape.com', meta=AUTOX_META)

    out = mw.process_request(req, spider)
    err = b'{"title":"No authentication token provided","type":"http://errors.xod.scrapinghub.com/unauthorized.html"}'
    resp = Response(out.url, request=out, body=err)
    with pytest.raises(AutoExtractError):
        mw.process_response(out, resp, spider)


def test_timeout():
    config = dict(MW_SETTINGS)
    # add a very low timeout - the middleware will ignore it
    config['AUTOEXTRACT_TIMEOUT'] = 1
    mw = _mock_mw(spider, config)
    req = Request('http://quotes.toscrape.com', meta=AUTOX_META)
    out = mw.process_request(req, spider)
    assert out is not None
    assert out.meta['download_timeout'] >= 180

    config['AUTOEXTRACT_TIMEOUT'] = 10000
    mw = _mock_mw(spider, config)
    req = Request('http://quotes.toscrape.com', meta=AUTOX_META)
    out = mw.process_request(req, spider)
    assert out is not None
    assert out.meta['download_timeout'] == 10000


def test_meta_extra():
    config = dict(MW_SETTINGS)
    mw = _mock_mw(spider, config)
    meta = deepcopy(AUTOX_META)
    meta['autoextract']['extra'] = {'some': 'stuff'}
    req = Request('http://quotes.toscrape.com', meta=meta)
    out = mw.process_request(req, spider)
    assert out is not None
    payload = json.loads(out.body.decode('utf8'))[0]
    assert payload['some'] == 'stuff'


def test_meta_headers():
    config = dict(MW_SETTINGS)
    mw = _mock_mw(spider, config)
    meta = deepcopy(AUTOX_META)
    meta['autoextract']['headers'] = {'UA': 'stuff'}
    req = Request('http://quotes.toscrape.com', meta=meta)
    out = mw.process_request(req, spider)
    assert out is not None
    assert out.headers['UA'].decode('utf8') == 'stuff'



import abc
import json

import sqlitedict
from autoextract.request import Request


class _Cache(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def fingerprint(cls, request: Request) -> str:
        return ''

    @abc.abstractmethod
    def __getitem__(self, fingerprint: str):
        pass

    @abc.abstractmethod
    def __setitem__(self, fingerprint: str, value) -> None:
        pass

    def close(self):
        pass


class DummyCache(_Cache):
    @classmethod
    def fingerprint(cls, request: Request) -> str:
        return ''

    def __getitem__(self, fingerprint: str):
        raise KeyError()

    def __setitem__(self, fingerprint: str, value) -> None:
        pass

    def __str__(self):
        return "no cache"


class AutoExtractCache(_Cache):
    def __init__(self, path):
        self.db = sqlitedict.SqliteDict(path, autocommit=True)

    @classmethod
    def fingerprint(cls, request: Request) -> str:
        return json.dumps(
            request.as_dict(),
            ensure_ascii=False,
            sort_keys=True
        )

    def __str__(self):
        return f"AutoExtractCache <{self.db.filename} | {len(self.db)} records>"

    def __getitem__(self, fingerprint: str):
        return self.db[fingerprint]

    def __setitem__(self, fingerprint: str, value) -> None:
        self.db[fingerprint] = value

    def close(self):
        self.db.close()
